{
  Yacc parse table construction.


  Copyright (c) 1990-92  Albert Graef <ag@muwiinfa.geschichte.uni-mainz.de>
  Copyright (C) 1998     Robert Zierer <rz1@stud.mw.tu-muenchen.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

$Revision: 1.3 $
$Date: 1999/06/29 19:00:51 $

}

{$undef DeallocTables}

unit YaccParseTable;

(* 3-1-91 AG *)
(* 4-22-98 RZ *)

interface


procedure parse_table;

  (* Constructs the parse table from the information in the state,
     transition and reduction table, and writes parse and rule table
     information to the output file.

     Rules never reduced are detected, and parsing conflicts resolved
     according to the usual disambiguting rules:

     - by default, shift/reduce conflicts are resolved in favour of
       shift, and reduce/reduce conflicts are resolved in favour of
       the rule appearing first in the grammar

     - in the presence of precedence information, shift/reduce conflicts
       are resolved as follows:
       - if the rule has higher precedence than the input symbol,
         reduce
       - if the input symbol has higher precedence than the rule,
         shift
       - if rule and input symbol have the same precedence, use
         associativity to resolve the conflict: if the symbol is
         left-associative, reduce; if right-associative, shift;
         if nonassociative, error.

     The default action for any state is error, unless the state
     only has a single reduce action, and no shift (or nonassoc-induced
     error) actions, in which case the default action is the reduction.
     An accept action is generated for the shift-endmarker action.

     If the verbose option is enabled, the parse_table routine also writes
     a readable listing of the generated parser to the .LST file, including
     descriptions of parse conflicts and rules never reduced.

     Parse table actions are encoded as follows:
     - positive: next state (shift or goto action)
     - negative: rule to reduce (reduce action)
     - 0: error (in default action table) or accept (in shift/reduce
          action table)

     The tables are written out as a collection of typed array constants:

     type YYARec = record { action record }
                     sym, act : Integer; { symbol and action }
                   end;
          YYRRec = record { rule record }
                     len, sym : Integer; { length and lhs symbol }
                   end;

     const

     yynacts   = ...; { number of parse table (shift and reduce) actions }
     yyngotos  = ...; { number of goto actions }
     yynstates = ...; { number of states }
     yynrules  = ...; { number of rules }

     yya : array [1..yynacts] of YYARec = ...;
       { shift and reduce actions }
     yyg : array [1..yyngotos] of YYARec = ...;
       { goto actions }
     yyd : array [0..yynstates-1] of Integer = ...;
       { default actions }
     yyal, yyah,
     yygl, yygh : array [0..yynstates-1] of Integer = ...;
       { offsets into action and goto table }

     yyr : array [1..yynrules] of YYRRec = ...;

  *)

var shift_reduce, reduce_reduce, never_reduced : Integer;
  (* number of parsing conflicts and unreduced rules detected during
     parse table generation *)

implementation

uses SysUtils, Classes,
     YaccBase, YaccTables, YaccLib, YaccMsgs, YaccMisc,
     {$ifdef VER80} YaccDlg16 {$else} {$ifdef WIN32} YaccDlg32 {$endif} {$endif}
{$ifdef LargeObjTables}
     , YaccTypes
{$endif}
     ;

var reduced : array [1..max_rules] of Boolean;

var yynacts, yyngotos, yynstates : Integer;
{$ifdef LargeObjTables}
    yyd,
    yyal, yyah, yygl, yygh : TDynLimArray_Integer; {array [0..max_states-1] of Integer;}
{$else}
    yyd : array [0..max_states-1] of Integer;
    yyal, yyah, yygl, yygh : array [0..max_states-1] of Integer;
{$endif}


function itemStr ( var item_set : ItemSet; i : Integer ) : String;
  (* returns print representation of item number i in item_set *)
  var str : String; j : Integer;
  begin
    with item_set, item[i], rule_table{$ifndef LargeObjTables}^{$endif}[rule_no]^ do
      begin
        str := pname(lhs_sym)+' :';
        for j := 1 to pos_no-1 do
          str := str+' '+pname(rhs_sym[j]);
        str := str+' _';
        for j := pos_no to rhs_len do
          str := str+' '+pname(rhs_sym[j]);
      end;
    itemStr := str;
  end(*itemStr*);

function ruleStr ( i : Integer ) : String;
  (* returns print representation of rule number i *)
  var str : String; j : Integer;
  begin
    with rule_table(*$ifndef LargeObjTables*)^(*$endif*)[i]^ do
      begin
        str := pname(lhs_sym)+' :';
        for j := 1 to rhs_len do
          str := str+' '+pname(rhs_sym[j]);
      end;
    ruleStr := str;
  end(*ruleStr*);

var
  RulestrStrings : TStringList;

procedure InitRulestrTable;
begin
  case RuleStringPlace of
    rs_Parser,
    rs_Text:
      begin
      RulestrStrings := TStringList.Create;
      with RulestrStrings do
        begin
        Clear;
        Sorted := false;
        Duplicates := dupAccept;
        end;
      end;
    rs_RC:
      OpenRCStrFile( yyRuleStrFile, RuleStrFilename );
    else ;
  end;
end; (*procedure InitRulestrTable*)

procedure FinalizeRulestrTable;
begin
  case RuleStringPlace of
    rs_Parser,
    rs_Text:
      RulestrStrings.Free;
    rs_RC:
      CloseRCStrFile( yyRuleStrFile );
    else ;
  end;
end; (*procedure FinalizeRulestrTable*)

procedure TuckInRuleString( Ident: word; var item_set : ItemSet; i : Integer );
var
  RuleStr: string;
begin (*procedure TuckInRuleString*)
  RuleStr := itemStr( item_set, i );
  case RuleStringPlace of
    rs_Parser,
    rs_Text:
      RulestrStrings.AddObject( RuleStr, Pointer(Ident) );
    rs_RC:
      WriteRCResStrContent( yyRuleStrFile, Ident, RuleStr );
    else ;
  end;
end; (*procedure TuckInRuleString*)

procedure emitRulestrings;

  function MakeString( s: string ): string;
  var i : integer;
  begin
    Result := '''';
    for i := 1 to Length( s ) do
      if s[i] = '''' then
        Result := Result + ''''''
      else
        Result := Result + s[i];
    Result := Result + '''';
  end;

var
  i,
  lastValue : longint;
  HasValueWritten : boolean;
begin
  case RuleStringPlace of
    rs_Parser:
      begin
      if not verbose then
        {no information available as not in verbose mode}
        writeln( yyout, '{no rulestrings available as grammar not in verbose mode processed}')
      else
        if RulestrStrings.Count <= 0 then
          raise EYaccException.Create( 'There were no Rulestrings' )
        else
          begin
          {emit yyrs}
          writeln( yyout, 'yyRules : array [1..' + IntToStr(RulestrStrings.Count) +
                         '] of string = (' );
          for i := 0 to RulestrStrings.Count-1 do
            begin
            if (i mod 50 = 0 )then YaccMsgs.beep;
            write( yyout, '{#' + IntToStr(i+1) + ': ' +
                          IntToStr( Longint( RulestrStrings.Objects[i] ) ) + '} ' );
            write( yyout, MakeString( RulestrStrings.Strings[i] ) );
            if i <> RulestrStrings.Count-1 then
              writeln( yyout, ',' )
            else
              begin
              Writeln( yyout );
              Writeln( yyout, ');' );
              Writeln( yyout );
              end;
            end;
          {emit yyrsl}
          {assuming Rulestrings are properly ordered; as they were collected by TuckInRuleString}
          i := Longint( RulestrStrings.Objects[ RulestrStrings.Count-1 ] ) -
               Longint( RulestrStrings.Objects[ 0 ] );
          {expecting they start with 0}
          writeln( yyout, 'yyRulesl : array [0..' + IntToStr(i) + '] of Integer = (' );
          lastValue := -1;
          for i := 0 to RulestrStrings.Count-1 do
            begin
            if (i mod 100 = 0 )then YaccMsgs.beep;
            HasValueWritten := false;
            if lastValue <> Longint( RulestrStrings.Objects[i] ) then
              begin
              write( yyout, '{' + IntToStr( Longint( RulestrStrings.Objects[i] ) ) + '} ' );
              {this value is first value;
               different counting: i is zero based, yyRules starts with 1}
              write( yyout, IntToStr( i + 1 ) );
              lastValue := Longint( RulestrStrings.Objects[i] );
              HasValueWritten := true;
              end;
            if i <> RulestrStrings.Count-1 then
              begin
              if HasValueWritten then
                writeln( yyout, ',' );
              end
            else
              begin
              Writeln( yyout );
              Writeln( yyout, ');' );
              Writeln( yyout );
              end;
            end;
          {emit yyrsh}
          i := Longint( RulestrStrings.Objects[ RulestrStrings.Count-1 ] ) -
               Longint( RulestrStrings.Objects[ 0 ] );
          writeln( yyout, 'yyRulesh : array [0..' + IntToStr(i) + '] of Integer = (' );
          {expecting they start with 0;
           and there should be no gaps in them (e.g. 3 and 5 are in, but not 4)}
          lastValue := 0;
          for i := 0 to RulestrStrings.Count-1 do
            begin
            if (i mod 100 = 0 )then YaccMsgs.beep;
            HasValueWritten := false;
            if lastValue <> Longint( RulestrStrings.Objects[i] ) then
              begin
              write( yyout, '{' + IntToStr( Longint( RulestrStrings.Objects[i] ) - 1 ) + '} ' );
              {previous value was last value;
               different counting: i is zero based, yyRules starts with 1}
              write( yyout, IntToStr( i ) );
              lastValue := Longint( RulestrStrings.Objects[i] );
              HasValueWritten := true;
              end;
            if HasValueWritten then
              writeln( yyout, ',' );
            if i = RulestrStrings.Count-1 then
              begin
              {last value}
              write( yyout, '{' + IntToStr(
                               Longint( RulestrStrings.Objects[ RulestrStrings.Count-1 ] ) -
                               Longint( RulestrStrings.Objects[ 0 ] )
                                  ) + '} ' );
              writeln( yyout, IntToStr( RulestrStrings.Count ) );
              Writeln( yyout, ');' );
              Writeln( yyout );
              end;
            end;
          end;
      end;
    rs_Text:
      if not verbose then
        {no information available as not in verbose mode}
        RenameIfExists( RuleStrFilename ) {instead of just deleting}
      else
        begin
        RenameIfExists( RuleStrFilename );
        AssignFile( yyRuleStrFile, RuleStrFilename );
        Rewrite( yyRuleStrFile );
        for i := 0 to RulestrStrings.Count-1 do
          Writeln( yyRuleStrFile, Longint( RulestrStrings.Objects[i] ),
                                  ': ', RulestrStrings.Strings[i] );
        CloseFile( yyRuleStrFile );
        end;
    rs_RC:
      ; {already emitted}
    else ;
  end;
end; (*procedure emitRulestrings*)

procedure build;

  (* build the parse table, resolve conflicts *)

  var

    i, j, k, s,
    n_errors,
    n_shifts,
    n_gotos,
    n_reductions,
    n_conflicts : Integer;
    html_conflicts : Integer;

  item_set : ItemSet;

      function GenerateHtmlRefForReduce( rule_no: Integer ): string;
      {RZ: Convert rule_no to state_no; rule_no is here passed as rule_no-1}
        {s should be state_no}
      (*var Res, _i_, yyr_sym : Integer;*)
      {Can't find a solution here; if you do please email me}
      begin
        Result := 'reduce&nbsp;' +
                  IntToStr(rule_no) ;
        {RZ: right ref here??; rather need a conversion to state_no}
      end;(*GenerateHtmlRefForReduce*)

var
  tmp : TransRec; z : word;
  begin
    YaccMsgs.beep;
    (* initialize: *)

    shift_reduce := 0; reduce_reduce := 0; never_reduced := 0;
    for i := 1 to n_rules do reduced[i] := false;

    (* traverse the state table: *)

    html_conflicts := 1;

    for s := 0 to n_states-1 do with state_table{$ifndef LargeObjTables}^{$endif}[s] do

      begin
        if (s mod 30 = 0 )then YaccMsgs.beep;
        if verbose then
          begin
            if GenHtml then
              begin
              writeln( yylst, '</OL><HR WIDTH="100%">');
              writeln( yylst, '<a name="' + IntToStr(s) + '">');
              end
            else
              begin
              writeln(yylst);
              end;
            writeln(yylst, 'state ', s, ':');
            if GenHtml then
              write( yylst, '</a><OL>' );
          end;

        (* Check shift and reduce actions, resolve conflicts.
           The number of error actions generated by nonassoc's is counted
           in n_errors, the number of conflicts reported in n_conflicts.
           Shift actions ruled out by disambiguating rules are flagged by
           setting the corresponding next_state to -1. *)

        n_errors := 0; n_conflicts := 0;

        for i := trans_lo to trans_hi do with trans_table{$ifndef LargeObjTables}^{$endif}[i] do
          if sym >= 0 then
            for j := redns_lo to redns_hi do with redn_table{$ifndef LargeObjTables}^{$endif}[j] do
              if member(sym, symset^) then
                if (sym_prec{$ifndef LargeObjTables}^{$endif}[sym]>0) and
                     (rule_prec{$ifndef LargeObjTables}^{$endif}[rule_no]>0) then
                  (* resolve conflict using precedence: *)
                  if rule_prec{$ifndef LargeObjTables}^{$endif}[rule_no] =
                       sym_prec{$ifndef LargeObjTables}^{$endif}[sym] then
                    case prec_table{$ifndef LargeObjTables}^{$endif}[
                           sym_prec{$ifndef LargeObjTables}^{$endif}[sym] ] of
                      left     : begin(* reduce *)
{$ifdef LargeObjTables}
                                 tmp := trans_table[i]; tmp.next_state := -1; trans_table[i] := tmp;
{$else}
                                 next_state := -1;
{$endif}
                                 end;
                      right    : (* shift *)
                                 exclude(symset^, sym);
                      nonassoc : (* error *)
                                 begin
                                   inc(n_errors);
{$ifdef LargeObjTables}
                                   tmp := trans_table[i]; tmp.next_state := -1; trans_table[i] := tmp;
{$else}
                                   next_state := -1;
{$endif}
                                   exclude(symset^, sym);
                                 end;
                    end
                  else if rule_prec{$ifndef LargeObjTables}^{$endif}[rule_no] >
                            sym_prec{$ifndef LargeObjTables}^{$endif}[sym] then begin
                    (* reduce *)
{$ifdef LargeObjTables}
                    tmp := trans_table[i]; tmp.next_state := -1; trans_table[i] := tmp;
{$else}
                    next_state := -1
{$endif}
                  end else
                    (* shift *)
                    exclude(symset^, sym)
                else
                  (* shift/reduce conflict: *)
                  begin
                    if verbose then
                      begin
                      if n_conflicts=0 then
                        begin
                        if GenHtml then
                          begin
                          writeln( yylst, '<a name="conflict'+IntToStr(html_conflicts)+'">');
                          writeln( yyLst, '<BR><FONT COLOR="#FF0000">*** conflicts:');
                          writeln( yyLst, '</a><BR><OL>');
                          end
                        else
                          begin
                          writeln(yylst);
                          writeln(yylst, tab, '*** conflicts:');
                          writeln(yylst);
                          end;
                        end;
                      if GenHtml then
                        begin
                        writeln( yyLst, '<BR>shift ' + IntToStr( next_state ) + ', ' +
                                        'reduce ' + IntToStr( rule_no-1) + ' on ' +
                                        pname(sym) + '' );
                        end
                      else
                        begin
                        writeln(yylst, tab,
                                       'shift ', next_state, ', ',
                                       'reduce ', rule_no-1, ' on ',
                                       pname(sym));
                        end;
                      end;
                    inc(n_conflicts); inc(shift_reduce);
                    inc(html_conflicts);
                    exclude(symset^, sym);
                  end;

        for i := redns_lo to redns_hi do
          for j := i+1 to redns_hi do with redn_table{$ifndef LargeObjTables}^{$endif}[j] do
            begin
              for k := 1 to size(symset^) do
                if member(symset^[k], redn_table{$ifndef LargeObjTables}^{$endif}[i].symset^) then
                  (* reduce/reduce conflict: *)
                  begin
                    if verbose then
                      begin
                      if n_conflicts=0 then
                        begin
                          if GenHtml then
                            begin
                            writeln( yylst, '<a name="conflict'+IntToStr(html_conflicts)+'">');
                            writeln( yyLst, '<BR><FONT COLOR="#FF0000">*** conflicts:');
                            writeln( yyLst, '</a><BR><OL>');
                            end
                          else
                            begin
                            writeln(yylst);
                            writeln(yylst, tab, '*** conflicts:');
                            writeln(yylst);
                            end;
                        end;
                      if GenHtml then
                        begin
                        writeln( yyLst, '<BR>reduce ' +
                                     IntToStr(
                                       redn_table{$ifndef LargeObjTables}^{$endif}[i].rule_no-1 ) +
                                     ', ' + 'reduce ' + IntToStr( rule_no-1 ) + ' on ' +
                                     pname(symset^[k]) );
                        end
                      else
                        begin
                        writeln(yylst, tab,
                                       'reduce ',
                                       redn_table{$ifndef LargeObjTables}^{$endif}[i].rule_no-1, ', ',
                                       'reduce ', rule_no-1, ' on ',
                                       pname(symset^[k]));
                        end;
                      end;
                    inc(n_conflicts); inc(reduce_reduce);
                    inc(html_conflicts);
                  end;
              setminus(symset^, redn_table{$ifndef LargeObjTables}^{$endif}[i].symset^);
            end;

        if GenHtml then
          if n_conflicts<>0 then
            begin
            writeln( yylst, '</OL></FONT>' );
            writeln( yyLst, '<a href="#conflict'+IntToStr(html_conflicts)+'">Next one!</a><BR>' )
            end
          else {n_conflicts=0}
            ;

        (* Count goto, shift and reduce actions to generate. *)

        n_gotos := 0; n_shifts := 0; n_reductions := 0;

        for i := trans_lo to trans_hi do with trans_table{$ifndef LargeObjTables}^{$endif}[i] do
          if next_state<>-1 then
            if sym<0 then
              inc(n_gotos)
            else
              inc(n_shifts);

        for i := redns_lo to redns_hi do with redn_table{$ifndef LargeObjTables}^{$endif}[i] do
          if size(symset^)>0 then
            inc(n_reductions);

        (* Determine default action. *)

        if (n_shifts+n_errors=0) and (n_reductions=1) then
          (* default action is the reduction *)
          with redn_table{$ifndef LargeObjTables}^{$endif}[redns_lo] do
            yyd[s] := -(rule_no-1)
        else
          (* default action is error *)
          yyd[s] := 0;

        (* Flag reduced rules. *)

        for i := redns_lo to redns_hi do
          with redn_table{$ifndef LargeObjTables}^{$endif}[i] do
            reduced[rule_no] := true;

        if verbose then

          begin

            (* List kernel items. *)

            if GenHtml then
              begin
              writeln( yyLst, '' );
              end
            else
              begin
              writeln(yylst);
              end;
            get_item_set(s, item_set);
            closure(item_set);
            sort_item_set(item_set);
            with item_set do
              begin
                for i := 1 to n_items do
                  with item[i], rule_table{$ifndef LargeObjTables}^{$endif}[rule_no]^ do
                    if (rule_no=1) or (pos_no>1) or (rhs_len=0) then
                      if pos_no>rhs_len then
                        begin
                        if RuleStrings then
                          TuckInRuleString( s, item_set, i );
                        if GenHtml then
                          begin
                          writeln( yyLst, '<BR>' + itemStr(item_set, i) + tab +
                                       '(' + IntToStr(rule_no-1) + ')' )
                          end
                        else
                          begin
                          writeln(yylst, tab,
                                         itemStr(item_set, i), tab,
                                         '(', rule_no-1, ')');
                          end
                        end
                      else
                        begin
                        if RuleStrings then
                          TuckInRuleString( s, item_set, i );
                        if GenHtml then
                          begin
                          writeln( yyLst, '<BR>' + itemStr(item_set, i) );
                          end
                        else
                          begin
                          writeln(yylst, tab, itemStr(item_set, i));
                          end;
                        end;
              end;

            (* List parse actions. *)

            (* shift, reduce and default actions: *)

            if GenHtml then
              begin
              writeln( yyLst, '<BR><BR><TABLE BORDER=0 COLS=3 WIDTH="25" >' );
              end;

            if (n_shifts+n_errors=0) and (n_reductions=1) then
              (* default action is the reduction *)
              with redn_table{$ifndef LargeObjTables}^{$endif}[redns_lo] do
                begin
                  if GenHtml then
                    begin
                    writeln( yyLst, '<TR><TD><BR></TD><TD><BR></TD><TD><BR></TD></TR>' );
                    write( yyLst, '<TR><TD>.</TD><TD><BR></TD>' );
                    writeln( yyLst, '<TD>' + GenerateHtmlRefForReduce(rule_no-1) +
                                    '</TD></TR>' );
                    end
                  else
                    begin
                    writeln(yylst);
                    writeln(yylst, tab, '.', tab, 'reduce ', rule_no-1 );
                    end;
                end
            else
              (* default action is error *)
              begin
                if GenHtml then
                  begin
                  writeln( yyLst, '<TR><TD><BR></TD><TD><BR></TD><TD><BR></TD></TR>' );
                  end
                else
                  begin
                  writeln(yylst);
                  end;
                for i := trans_lo to trans_hi do with trans_table{$ifndef LargeObjTables}^{$endif}[i] do
                  if next_state<>-1 then
                    if sym=0 then
                      (* accept action *)
                      if GenHtml then
                        begin
                        writeln( yyLst, '<TR><TD>' + pname(sym) +
                                        '</TD><TD><BR></TD><TD>accept</TD></TR>')
                        end
                      else
                        begin
                        writeln(yylst, tab, pname(sym), tab, 'accept')
                        end
                    else if sym>0 then
                      (* shift action *)
                      if GenHtml then
                        begin
                        writeln( yyLst, '<TR><TD>' + pname(sym) +
                                        '</TD><TD><BR></TD><TD><a href="#' +
                                        IntToStr(next_state) +
                                        '">shift&nbsp;' + IntToStr(next_state) +
                                        '</a></TD></TR>' );
                        end
                      else
                        begin
                        writeln(yylst, tab, pname(sym), tab, 'shift ', next_state);
                        end;
                for i := redns_lo to redns_hi do
                  with redn_table{$ifndef LargeObjTables}^{$endif}[i] do
                    for j := 1 to size(symset^) do
                      (* reduce action *)
                      if GenHtml then
                        begin
                        write( yyLst, '<TR><TD>' + pname(symset^[j]) +
                                     '</TD><TD><BR></TD>' );
                        writeln( yyLst, '<TD>' + GenerateHtmlRefForReduce(rule_no-1) +
                                        '</TD></TR>' );
                        end
                      else
                        begin
                        writeln(yylst, tab, pname(symset^[j]), tab, 'reduce ',
                                       rule_no-1);
                        end;
                (* error action *)
                if GenHtml then
                  begin
                  writeln( yyLst, '<TR><TD><BR></TD><TD><BR></TD><TD><BR></TD></TR>' );
                  writeln( yyLst, '<TR><TD>.</TD><TD><BR></TD><TD>error</TD></TR>' );
                  end
                else
                  begin
                  writeln(yylst);
                  writeln(yylst, tab, '.', tab, 'error');
                  end;
              end;

            (* goto actions: *)

            if n_gotos>0 then
              begin
                if GenHtml then
                  writeln( yyLst, '<TR><TD><BR></TD><TD><BR></TD><TD><BR></TD></TR>' )
                else
                  writeln(yylst);
                for i := trans_lo to trans_hi do with trans_table{$ifndef LargeObjTables}^{$endif}[i] do
                  if sym<0 then
                    if GenHtml then
                      begin
                      writeln( yyLst, '<TR><TD>' + pname(sym) + '</TD><TD><BR></TD><TD>' +
                                      '<a href="#' + IntToStr(next_state) + '">goto&nbsp;' +
                                      IntToStr(next_state) + '</a></TD></TR>');
                      end
                    else
                      begin
                      writeln(yylst, tab, pname(sym), tab, 'goto ', next_state);
                      end;
              end;

            if GenHtml then
              writeln( yyLst, '</TABLE>' );
          end;

      end;

    if GenHtml then
      writeln( yylst, '</OL><HR WIDTH="100%">');

    for i := 2 to n_rules do
      if not reduced[i] then
        inc(never_reduced);

    if verbose then
      begin

      if GenHtml then
        begin
        writeln( yyLst, '<BR>' );
        end;

      if shift_reduce>0 then
        if GenHtml then
          begin
          writeln( yyLst, '<BR><FONT COLOR="#FF0000">' + IntToStr(shift_reduce) +
                               ' shift/reduce conflicts.</Font>');
          GiveOutLine( omScreen, IntToStr(shift_reduce) + ' shift/reduce conflicts.');
          end
        else
          GiveOutLine( omBoth, IntToStr(shift_reduce) + ' shift/reduce conflicts.');

      if reduce_reduce>0 then
        if GenHtml then
          begin
          writeln( yyLst, '<BR><FONT COLOR="#FF0000">' + IntToStr(reduce_reduce) +
                                 ' reduce/reduce conflicts.</Font>');
          GiveOutLine( omScreen, IntToStr(reduce_reduce) + ' reduce/reduce conflicts.');
          end
        else
          GiveOutLine( omBoth, IntToStr(reduce_reduce) + ' reduce/reduce conflicts.');

      if GenHtml and ( ( reduce_reduce>0 ) or ( shift_reduce>0 ) ) then
        begin
        write( yyLst, '<OL><a name="conflict'+IntToStr(html_conflicts)+'"></a>' );
        write( yylst, '<a href="#conflict'+IntToStr(1)+'">Where they are</a></OL>' );
        end;

      if never_reduced>0 then
        if GenHtml then
          begin
          write( yyLst, '<FONT COLOR="#FF0000">' + IntToStr(never_reduced) +
                                   ' rules never reduced.</Font>');
          GiveOutLine( omScreen, IntToStr(never_reduced) + ' rules never reduced.');
          end
        else
          GiveOutLine( omBoth, IntToStr(never_reduced) + ' rules never reduced.');
      end;

    (* report rules never reduced: *)

    if (never_reduced>0) and verbose then
      begin
      GiveOutLine( omBoth, '');
      if GenHtml then
        begin
        writeln( yyLst, '<BR><FONT COLOR="#FF0000">*** rules never reduced:<OL>' );
        GiveOutLine( omScreen, '*** rules never reduced:' );
        end
      else
        GiveOutLine( omBoth, '*** rules never reduced:' );
      for i := 2 to n_rules do
        if not reduced[i] then
          begin
          if GenHtml then
            begin
            writeln( yyLst, ruleStr(i) + '&nbsp;&nbsp;&nbsp;' +
                                         '(' + IntToStr(i-1) + ')<BR>');
            GiveOutLine( omScreen, '');
            GiveOutLine( omScreen, ruleStr(i) + tab + '(' + IntToStr(i-1) + ')');
            end
          else
            begin
            GiveOutLine( omBoth, '');
            GiveOutLine( omBoth, ruleStr(i) + tab + '(' + IntToStr(i-1) + ')');
            end;
          end;
      if GenHtml then
        writeln( yyLst, '</OL></FONT>')
      end;

  end(*build*);

procedure counters;

  (* initialize counters and offsets *)

  var s, i : Integer;

  begin
    YaccMsgs.beep;

    yynstates := n_states; yynacts := 0; yyngotos := 0;

    for s := 0 to n_states-1 do with state_table{$ifndef LargeObjTables}^{$endif}[s] do
      begin
        if (s mod 40 = 0 ) then YaccMsgs.beep;
        yyal[s] := yynacts+1; yygl[s] := yyngotos+1;
        if yyd[s]=0 then
          begin
            for i := trans_lo to trans_hi do with trans_table{$ifndef LargeObjTables}^{$endif}[i] do
              if (sym>=0) and (next_state<>-1) then
                inc(yynacts);
            for i := redns_lo to redns_hi do with redn_table{$ifndef LargeObjTables}^{$endif}[i] do
              inc(yynacts, size(symset^));
          end;
        for i := trans_lo to trans_hi do with trans_table{$ifndef LargeObjTables}^{$endif}[i] do
          if sym<0 then
            inc(yyngotos);
        yyah[s] := yynacts; yygh[s] := yyngotos;
      end;

  end(*counters*);

procedure tables;

  procedure NeedSeparator( a, b: integer );
    begin
      if a < b then
        case TableType of
          ttNormal :
            write(yyout, ',');
          ttExtern :
            ;
          ttResource :
            ;
          else
            fatal(internal_error)
        end;
      case TableType of
        ttNormal :
          writeln(yyout);
        ttExtern :
          ;
        ttResource :
          ;
        else
          fatal(internal_error)
      end;
    end;

  var yyTableOut : file of integer;

  procedure EndSeparator;
    begin
      case TableType of
        ttNormal :
          writeln(yyout, ');');
        ttExtern :
          CloseFile( yyTableOut );
        ttResource :
          ;
        else
          fatal(internal_error)
      end;
    end;

  (* write tables to output file *)

  var TableValue1, TableValue2 : integer;
  type YYARec = record
                sym, act : Integer;
                end;
       YYRRec = record
                len, sym : Integer;
                end;

  var s, i, j, count : Integer;

  begin
    YaccMsgs.beep;

    case TableType of
      ttNormal,
      ttExtern :
        ;
      ttResource :
        begin
        RenameIfExists( ChangeFileExt( pasfilename, '.RES' ) );
        AssignFile( yyResTable, ChangeFileExt( pasfilename, '.RES' ) );
        Rewrite( yyResTable );
        end;
      else
        fatal(internal_error)
    end;

    writeln(yyout);
    writeln(yyout, 'type YYARec = record');
    writeln(yyout, '                sym, act : Integer;');
    writeln(yyout, '              end;');
    writeln(yyout, '     YYRRec = record');
    writeln(yyout, '                len, sym : Integer;');
    writeln(yyout, '              end;');
    writeln(yyout);
    writeln(yyout, 'const');

    (* counters: *)

    writeln(yyout);
    case TableType of
      ttNormal,
      ttExtern :
        begin
        writeln(yyout, 'yynacts   = ', yynacts, ';');
        writeln(yyout, 'yyngotos  = ', yyngotos, ';');
        writeln(yyout, 'yynstates = ', yynstates, ';');
        writeln(yyout, 'yynrules  = ', n_rules-1, ';');
        end;
      ttResource :
        begin
        writeln(yyout, 'yynacts  : word = 0;');
        writeln(yyout, 'yyngotos : word = 0;');
        writeln(yyout, 'yynstates: word = 0;');
        writeln(yyout, 'yynrules : word = 0;');
        writeln(yyout, '{$R-}' );
        WriteRCResHeader( yyResTable, ResTablePrefix+'TABLESIZES', 4*SizeOf(word) );
        WriteRCResContent( yyResTable, [ integer(yynacts), integer(yyngotos),
                                         integer(yynstates), integer(n_rules-1) ] );
        end;
      else
        fatal(internal_error)
    end;

    try

      (* shift/reduce table: *)

      writeln(yyout);
      case TableType of
        ttNormal :
          writeln(yyout, 'yya : array [1..yynacts] of YYARec = (');
        ttExtern :
          begin
          writeln(yyout, 'type yya_type = array [1..yynacts] of YYARec;' );
          writeln(yyout, 'const yya : ^yya_type = nil;');
          RenameIfExists( ExtractFilePath(pasfilename)+'yya.dat' );
          AssignFile( yyTableOut, ExtractFilePath(pasfilename)+'yya.dat');
          Rewrite( yyTableOut );
          end;
        ttResource :
          begin
          writeln(yyout, 'type yya_type = array [1..1{yynacts}] of YYARec;' );
          writeln(yyout, 'const yya : ^yya_type = nil;');
          WriteRCResHeader( yyResTable, ResTablePrefix+'YYA_DAT', yynacts*SizeOf(YYARec) );
          end;
        else
          fatal(internal_error)
      end;
      count := 0;
      for s := 0 to n_states-1 do with state_table{$ifndef LargeObjTables}^{$endif}[s] do
        begin
          if (s mod 30 = 0 ) then YaccMsgs.beep;
          case TableType of
            ttNormal :
              writeln(yyout, '{ ', s, ': }');
            ttExtern :
              ;
            ttResource :
              ;
            else
              fatal(internal_error)
          end;
          if yyd[s]=0 then
            begin
              for i := trans_lo to trans_hi do with trans_table{$ifndef LargeObjTables}^{$endif}[i] do
                if (next_state<>-1) and (sym>=0) then
                  begin
                    inc(count);
                    if sym=0 then begin
                      case TableType of
                        ttNormal :
                          write(yyout, '  ( sym: 0; act: 0 )');
                        ttExtern :
                          begin
                          TableValue1 := 0;
                          TableValue2 := 0;
                          write(yyTableOut, TableValue1, TableValue2 );
                          end;
                        ttResource :
                          WriteRCResContent(yyResTable, [0,0]);
                        else
                          fatal(internal_error)
                      end;
                    end else begin
                      case TableType of
                        ttNormal :
                          write(yyout, '  ( sym: ', sym, '; act: ', next_state, ' )');
                        ttExtern :
                          begin
                          TableValue1 := sym;
                          TableValue2 := next_state;
                          write(yyTableOut, TableValue1, TableValue2 );
                          end;
                        ttResource :
                          WriteRCResContent(yyResTable, [sym,next_state] );
                        else
                          fatal(internal_error)
                      end;
                    end;
                    NeedSeparator( count, yynacts );
                  end;
              for i := redns_lo to redns_hi do with redn_table{$ifndef LargeObjTables}^{$endif}[i] do
                for j := 1 to size(symset^) do
                  begin
                    inc(count);
                    case TableType of
                      ttNormal :
                        write( yyout, '  ( sym: ', symset^[j],
                                      '; act: ', -(rule_no-1), ' )' );
                      ttExtern :
                        begin
                        TableValue1 := symset^[j];
                        TableValue2 := -(rule_no-1);
                        write(yyTableOut, TableValue1, TableValue2 );
                        end;
                      ttResource :
                        WriteRCResContent(yyResTable, [symset^[j],-(rule_no-1)]);
                      else
                        fatal(internal_error)
                    end;
                    NeedSeparator( count, yynacts );
                  end;
          end;
        end;
      EndSeparator;

      (* goto table: *)

      writeln(yyout);
      case TableType of
        ttNormal :
          writeln(yyout, 'yyg : array [1..yyngotos] of YYARec = (');
        ttExtern :
          begin
          writeln(yyout, 'type yyg_type = array [1..yyngotos] of YYARec;' );
          writeln(yyout, 'const yyg : ^yyg_type = nil;');
          RenameIfExists( ExtractFilePath(pasfilename)+'yyg.dat' );
          AssignFile( yyTableOut, ExtractFilePath(pasfilename)+'yyg.dat');
          Rewrite( yyTableOut );
          end;
        ttResource :
          begin
          writeln(yyout, 'type yyg_type = array [1..1{yyngotos}] of YYARec;' );
          writeln(yyout, 'const yyg : ^yyg_type = nil;');
          WriteRCResHeader( yyResTable, ResTablePrefix+'YYG_DAT', yyngotos*SizeOf(YYARec) );
          end;
        else
          fatal(internal_error)
      end;
      count := 0;
      for s := 0 to n_states-1 do with state_table{$ifndef LargeObjTables}^{$endif}[s] do
        begin
          case TableType of
            ttNormal :
              writeln(yyout, '{ ', s, ': }');
            ttExtern :
              ;
            ttResource :
              ;
            else
              fatal(internal_error)
          end;
          for i := trans_lo to trans_hi do with trans_table{$ifndef LargeObjTables}^{$endif}[i] do
            if sym<0 then
              begin
                inc(count);
                case TableType of
                  ttNormal :
                    write(yyout, '  ( sym: ', sym, '; act: ', next_state, ' )');
                  ttExtern :
                    begin
                    TableValue1 := sym;
                    TableValue2 := next_state;
                    write(yyTableOut, TableValue1, TableValue2 );
                    end;
                  ttResource :
                    WriteRCResContent(yyResTable, [sym,next_state]);
                  else
                    fatal(internal_error)
                end;
                NeedSeparator( count, yyngotos );
              end;
        end;
      EndSeparator;

      (* default action table: *)

      writeln(yyout);
      case TableType of
        ttNormal :
          writeln(yyout, 'yyd : array [0..yynstates-1] of Integer = (');
        ttExtern :
          begin
          writeln(yyout, 'type yyd_type = array [0..yynstates-1] of Integer;' );
          writeln(yyout, 'const yyd : ^yyd_type = nil;');
          RenameIfExists( ExtractFilePath(pasfilename)+'yyd.dat' );
          AssignFile( yyTableOut, ExtractFilePath(pasfilename)+'yyd.dat');
          Rewrite( yyTableOut );
          end;
        ttResource :
          begin
          writeln(yyout, 'type yyd_type = array [0..0{yynstates-1}] of Integer;' );
          writeln(yyout, 'const yyd : ^yyd_type = nil;');
          WriteRCResHeader( yyResTable, ResTablePrefix+'YYD_DAT', n_states*SizeOf(integer) );
          end;
        else
          fatal(internal_error)
      end;
      for s := 0 to n_states-1 do
        begin
          case TableType of
            ttNormal :
              write(yyout, '{ ', s, ': } ', yyd[s]);
            ttExtern :
              begin
              TableValue1 := yyd[s];
              write(yyTableOut, TableValue1 );
              end;
            ttResource :
              WriteRCResContent(yyResTable, [yyd[s]]);
            else
              fatal(internal_error)
          end;
          NeedSeparator( s, n_states-1 );
        end;
      EndSeparator;

      (* offset tables: *)

      writeln(yyout);
      case TableType of
        ttNormal :
          writeln(yyout, 'yyal : array [0..yynstates-1] of Integer = (');
        ttExtern :
          begin
          writeln(yyout, 'type yyal_type = array [0..yynstates-1] of Integer;' );
          writeln(yyout, 'const yyal : ^yyal_type = nil;');
          RenameIfExists( ExtractFilePath(pasfilename)+'yyal.dat' );
          AssignFile( yyTableOut, ExtractFilePath(pasfilename)+'yyal.dat');
          Rewrite( yyTableOut );
          end;
        ttResource :
          begin
          writeln(yyout, 'type yyal_type = array [0..0{yynstates-1}] of Integer;' );
          writeln(yyout, 'const yyal : ^yyal_type = nil;');
          WriteRCResHeader( yyResTable, ResTablePrefix+'YYAL_DAT', n_states*SizeOf(integer) );
          end;
        else
          fatal(internal_error)
      end;
      for s := 0 to n_states-1 do
        begin
          case TableType of
            ttNormal :
              write(yyout, '{ ', s, ': } ', yyal[s]);
            ttExtern :
              begin
              TableValue1 := yyal[s];
              write(yyTableOut, TableValue1 );
              end;
            ttResource :
              WriteRCResContent(yyResTable, [yyal[s]]);
            else
              fatal(internal_error)
          end;
          NeedSeparator( s, n_states-1 );
        end;
      EndSeparator;

      writeln(yyout);
      case TableType of
        ttNormal :
          writeln(yyout, 'yyah : array [0..yynstates-1] of Integer = (');
        ttExtern :
          begin
          writeln(yyout, 'type yyah_type = array [0..yynstates-1] of Integer;' );
          writeln(yyout, 'const yyah : ^yyah_type = nil;');
          RenameIfExists( ExtractFilePath(pasfilename)+'yyah.dat' );
          AssignFile( yyTableOut, ExtractFilePath(pasfilename)+'yyah.dat');
          Rewrite( yyTableOut );
          end;
        ttResource :
          begin
          writeln(yyout, 'type yyah_type = array [0..0{yynstates-1}] of Integer;' );
          writeln(yyout, 'const yyah : ^yyah_type = nil;');
          WriteRCResHeader( yyResTable, ResTablePrefix+'YYAH_DAT', n_states*SizeOf(integer) );
          end;
        else
          fatal(internal_error)
      end;
      for s := 0 to n_states-1 do
        begin
          case TableType of
            ttNormal :
              write(yyout, '{ ', s, ': } ', yyah[s]);
            ttExtern :
              begin
              TableValue1 := yyah[s];
              write(yyTableOut, TableValue1 );
              end;
            ttResource :
              WriteRCResContent(yyResTable, [yyah[s]]);
            else
              fatal(internal_error)
          end;
          NeedSeparator( s, n_states-1 );
        end;
      EndSeparator;


      writeln(yyout);
      case TableType of
        ttNormal :
          writeln(yyout, 'yygl : array [0..yynstates-1] of Integer = (');
        ttExtern :
          begin
          writeln(yyout, 'type yygl_type = array [0..yynstates-1] of Integer;' );
          writeln(yyout, 'const yygl : ^yygl_type = nil;');
          RenameIfExists( ExtractFilePath(pasfilename)+'yygl.dat' );
          AssignFile( yyTableOut, ExtractFilePath(pasfilename)+'yygl.dat');
          Rewrite( yyTableOut );
          end;
        ttResource :
          begin
          writeln(yyout, 'type yygl_type = array [0..0{yynstates-1}] of Integer;' );
          writeln(yyout, 'const yygl : ^yygl_type = nil;');
          WriteRCResHeader( yyResTable, ResTablePrefix+'YYGL_DAT', n_states*SizeOf(integer) );
          end;
        else
          fatal(internal_error)
      end;
      for s := 0 to n_states-1 do
        begin
          case TableType of
            ttNormal :
              write(yyout, '{ ', s, ': } ', yygl[s]);
            ttExtern :
              begin
              TableValue1 := yygl[s];
              write(yyTableOut, TableValue1 );
              end;
            ttResource :
              WriteRCResContent(yyResTable, [yygl[s]]);
            else
              fatal(internal_error)
          end;
          NeedSeparator( s, n_states-1 );
        end;
      EndSeparator;

      writeln(yyout);
      case TableType of
        ttNormal :
          writeln(yyout, 'yygh : array [0..yynstates-1] of Integer = (');
        ttExtern :
          begin
          writeln(yyout, 'type yygh_type = array [0..yynstates-1] of Integer;' );
          writeln(yyout, 'const yygh : ^yygh_type = nil;');
          RenameIfExists( ExtractFilePath(pasfilename)+'yygh.dat' );
          AssignFile( yyTableOut, ExtractFilePath(pasfilename)+'yygh.dat');
          Rewrite( yyTableOut );
          end;
        ttResource :
          begin
          writeln(yyout, 'type yygh_type = array [0..0{yynstates-1}] of Integer;' );
          writeln(yyout, 'const yygh : ^yygh_type = nil;');
          WriteRCResHeader( yyResTable, ResTablePrefix+'YYGH_DAT', n_states*SizeOf(integer) );
          end;
        else
          fatal(internal_error)
      end;
      for s := 0 to n_states-1 do
        begin
          case TableType of
            ttNormal :
              write(yyout, '{ ', s, ': } ', yygh[s]);
            ttExtern :
              begin
              TableValue1 := yygh[s];
              write(yyTableOut, TableValue1 );
              end;
            ttResource :
              WriteRCResContent(yyResTable, [yygh[s]]);
            else
              fatal(internal_error)
          end;
          NeedSeparator( s, n_states-1 );
        end;
      EndSeparator;

      (* rule table: *)

      writeln(yyout);
      case TableType of
        ttNormal :
          writeln(yyout, 'yyr : array [1..yynrules] of YYRRec = (');
        ttExtern :
          begin
          writeln(yyout, 'type yyr_type = array [1..yynrules] of YYRRec;' );
          writeln(yyout, 'const yyr : ^yyr_type = nil;');
          RenameIfExists( ExtractFilePath(pasfilename)+'yyr.dat' );
          AssignFile( yyTableOut, ExtractFilePath(pasfilename)+'yyr.dat');
          Rewrite( yyTableOut );
          end;
        ttResource :
          begin
          writeln(yyout, 'type yyr_type = array [1..1{yynrules}] of YYRRec;' );
          writeln(yyout, 'const yyr : ^yyr_type = nil;');
          WriteRCResHeader( yyResTable, ResTablePrefix+'YYR_DAT', (n_rules-1)*SizeOf(YYRRec) );
          end;
        else
          fatal(internal_error)
      end;
      for i := 2 to n_rules do with rule_table{$ifndef LargeObjTables}^{$endif}[i]^ do
        begin
          case TableType of
            ttNormal :
              write(yyout, '{ ', i-1, ': } ', '( len: ', rhs_len,
                           '; sym: ', lhs_sym, ' )');
            ttExtern :
              begin
              TableValue1 := rhs_len;
              TableValue2 := lhs_sym;
              write(yyTableOut, TableValue1, TableValue2 );
              end;
            ttResource :
              WriteRCResContent(yyResTable, [rhs_len,lhs_sym]);
            else
              fatal(internal_error)
          end;
          NeedSeparator( i, n_rules );
        end;
      EndSeparator;

      case TableType of
        ttNormal :
          writeln(yyout);
        ttExtern :
          ;
        ttResource :
          ;
        else
          fatal(internal_error)
      end;

    except
      case TableType of
        ttNormal :
          ;
        ttExtern,
        ttResource:
          CloseFile( yyTableOut );
        else
          fatal(internal_error)
      end;
    end;

  end(*tables*);

procedure parse_table;
  begin
    if RuleStrings then
      InitRulestrTable;

    build; counters; tables;

    if RuleStrings then
      begin
      emitRulestrings;
      FinalizeRulestrTable;
      end;
  end(*parse_table*);


{************************************************************************************}
{************************************************************************************}
procedure DeallocateTables;
begin
{$ifdef DeallocTables}
    yyd.Free;
    yyal.Free;
    yyah.Free;
    yygl.Free;
    yygh.Free;
{$endif}
end; (*proc DeallocateTables*)

var OldExitProc: pointer;

{$F+}
procedure NewExitProc; far;
{$F-}
begin
  try
    DeallocateTables;
  finally
    ExitProc := OldExitProc;
  end;
end;

initialization
{$ifdef LargeObjTables}
{$ifdef DeallocTables}
  OldExitProc := ExitProc;
  ExitProc := @NewExitProc;
{$endif}
      {yyd = array [0..max_states-1] of Integer;}
    yyd := TDynLimArray_Integer.Create;
    yyd.Dim( 0, max_states-1 );
      {yyal = array [0..max_states-1] of Integer;}
    yyal := TDynLimArray_Integer.Create;
    yyal.Dim( 0, max_states-1 );
      {yyah = array [0..max_states-1] of Integer;}
    yyah := TDynLimArray_Integer.Create;
    yyah.Dim( 0, max_states-1 );
      {yygl = array [0..max_states-1] of Integer;}
    yygl := TDynLimArray_Integer.Create;
    yygl.Dim( 0, max_states-1 );
      {yygh = array [0..max_states-1] of Integer;}
    yygh := TDynLimArray_Integer.Create;
    yygh.Dim( 0, max_states-1 );
{$endif}
end(*YaccParseTable*).

