{
  This module collects the various tables used by the Yacc program:
   - the symbol table
   - the rule table
   - the precedence table
   - the closure table
   - the LALR state, item, transition and reduction table
   Note: All tables are allocated dynamically (at initialization time)
   because of the 64KB static data limit.


  Copyright (c) 1990-92  Albert Graef <ag@muwiinfa.geschichte.uni-mainz.de>
  Copyright (C) 1998     Robert Zierer <rz1@stud.mw.tu-muenchen.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

$Revision: 1.3 $
$Date: 1999/06/29 19:00:52 $

}

unit YaccTables;

(* 2-16-91 AG *)
(* 4-22-98 RZ; Bug fixed in function n_nt_rules *)

interface
{$X+}
uses YaccBase
{$ifdef LargeObjTables}
     , YaccTypes
{$endif}
     ;

var max_bytes : LongInt;
  (* available memory *)

function n_bytes : LongInt;
  (* memory actually used *)

const

(* Prime-table for max_keys: *)
(*
Primes : array[1..10] of word = ( 997, 1999, 2999, 3989, 4999, 5987, 6997, 7993, 8999, 9973 );
         (* last two exceed SymTable to over 65K *)

(* Maximum table sizes: *)

max_keys    {: word} =  997;  (* size of hash symbol table (prime number!)   *)
max_nts     {: word} =  2000;{300}  (* maximum number of nonterminals              *)
max_lits    {: word} =  max_nts + 256;  (* number of literals ( (max_nts) 300+256)       *)
max_rules   {: word} =  max_nts + 1;  (* number of rules ( (max_nts) 300+1)                     *)
max_types   {: word} =  100;  (* number of type tags                         *)
max_prec    {: word} =   50;  (* maximum precedence level                    *)
max_states  {: word} = 1600;{600}  (* number of LR(0) states                      *)
max_items   {: word} = 9000;{3000}  (* number of items                             *)
max_trans   {: word} = 9000;{3000}  (* number of transitions                       *)
max_redns   {: word} = 1200;  (* number of reductions                        *)

max_rule_len  {: word} =   64; {64}  (* maximum length of rules                     *)
max_set_items {: word} =  200; {100} (* maximum number of items in an item set      *)

var

(* Actual table sizes: *)

n_nts            : Integer;
n_lits           : Integer;
n_rules          : Integer;
n_types          : Integer;
n_prec           : Integer;
n_states         : Integer;
n_items          : Integer;
n_trans          : Integer;
n_redns          : Integer;

type

(* Table data structures: *)

(* Symbol table: The symbol table consists of a hash table which stores
   print names and internal symbol numbers, and a key table which stores,
   for each internal symbol number, the corresponding hash key. *)

SymRec = record
           pname  : StrPtr;
             (* print name; empty entries are denoted by pname=nil *)
           deff   : Boolean;
             (* flag denoting whether symbol is already defined *)
           sym    : Integer;
             (* internal symbol number (0 or positive: literal symbols
                (literal characters have symbol numbers 1 thru 255);
                negative: nonterminal symbols; 0 denotes endmarker,
                -1 augmented start nonterminal, 256 is reserved for
                error token; note that the predefined symbols except
                the error literal are not actually stored in the symbol
                table; the error symbol is entered at initialization
                s.t. it always has number 256) *)
         end;

{=================================================================================}

{$ifdef LargeObjTables}
  {a redimensionable array with elements of type YaccTables.SymRec }
  TDynLimArray_SymRec = class(TDynLimArrayRes)
     protected
       procedure SetElement( Index: longint; value: SymRec );
       function GetElement( Index: longint ): SymRec;
     public
       property Element[ Index: longint ]: SymRec read GetElement write SetElement; default;
       constructor Create;
     end;
{$else}
SymTable = array [1..max_keys] of SymRec;
{$endif}

{$ifdef LargeObjTables}
{TDynLimArray_Integer in YaccTypes}
{$else}
SymKeyTable = array [-max_nts..max_lits-1] of Integer;
{$endif}

  (* hash keys for nonterminal and literal symbols *)

(* Rule table: the rule table consists of an array storing the rules
   sequentially in the order in which they appear in the source grammar;
   a rule no.s table which is used to sort rules w.r.t. left-hand side
   nonterminals (after the rule table has been constructed and sorted, all
   references to rules are done indirectly via the rule_no array s.t. the
   rules for each nonterminal can be accessed easily); and an offset table
   which stores, for each nonterminal, the corresponding first and last
   index in the rule no.s table. *)

RuleRec = record
            lhs_sym : Integer; (* lhs nonterminal *)
            rhs_len : Integer; (* length of rhs *)
            rhs_sym : array [1..max_rule_len] of Integer;
              (* rhs symbols *)
          end;

RuleRecPtr = ^RuleRec;

{=================================================================================}

{$ifdef LargeObjTables}
  {a redimensionable array with elements of type YaccTables.RuleRecPtr }
  TDynLimArray_RuleRecPtr = class(TDynLimArrayRes)
     protected
       procedure SetElement( Index: longint; value: RuleRecPtr );
       function GetElement( Index: longint ): RuleRecPtr;
     public
       property Element[ Index: longint ]: RuleRecPtr read GetElement write SetElement; default;
       constructor Create;
     end;
{$else}
RuleTable = array [1..max_rules] of RuleRecPtr;
{$endif}

{$ifdef LargeObjTables}
{TDynLimArray_Integer in YaccTypes}
{$else}
RuleNoTable = array [1..max_rules] of Integer;
{$endif}

RuleOffsRec = record
                rule_lo, rule_hi : Integer;
              end;

{=================================================================================}

{$ifdef LargeObjTables}
  {a redimensionable array with elements of type YaccTables.RuleOffsRec }
  TDynLimArray_RuleOffsRec = class(TDynLimArrayRes)
     protected
       procedure SetElement( Index: longint; value: RuleOffsRec );
       function GetElement( Index: longint ): RuleOffsRec;
     public
       property Element[ Index: longint ]: RuleOffsRec read GetElement write SetElement; default;
       constructor Create;
     end;
{$else}
RuleOffsTable = array [1..max_nts] of RuleOffsRec;
{$endif}

(* Symbol type table: The symbol type table stores the types associated
   with the nonterminal and terminal grammar symbols (0 if none). *)

{$ifdef LargeObjTables}
{TDynLimArray_Integer in YaccTypes}
{$else}
TypeTable    = array [1..max_types] of Integer;
{$endif}

  (* types declared in the definitions section *)

{$ifdef LargeObjTables}
{TDynLimArray_Integer in YaccTypes}
{$else}
SymTypeTable = array [-max_nts..max_lits-1] of Integer;
{$endif}

  (* symbol types *)

(* Precedence table: The precedence table stores the type of each
   precedence level (left, right, nonassoc) and, for each literal
   symbol and grammar rule, the assigned precedence level (precedence
   level 0 if none). *)

PrecType = ( left, right, nonassoc );

{=================================================================================}

{$ifdef LargeObjTables}
  {a redimensionable array with elements of type YaccTables.PrecType }
  TDynLimArray_PrecType = class(TDynLimArrayRes)
     protected
       procedure SetElement( Index: longint; value: PrecType );
       function GetElement( Index: longint ): PrecType;
     public
       property Element[ Index: longint ]: PrecType read GetElement write SetElement; default;
       constructor Create;
     end;
{$else}
PrecTable = array [1..max_prec] of PrecType;
{$endif}

{$ifdef LargeObjTables}
{TDynLimArray_Integer in YaccTypes}
{$else}
SymPrecTable = array [0..max_lits-1] of Integer;
{$endif}

{$ifdef LargeObjTables}
{TDynLimArray_Integer in YaccTypes}
{$else}
RulePrecTable = array [1..max_rules] of Integer;
{$endif}

(* Closure and first symbols table: The closure table stores, for each
   nonterminal X, the set of those nonterminals Y for which there is a
   rightmost derivation X =>+ Y ... . Similarly, the first set table
   stores, for each nonterminal X, the set of literals a for which there
   is a derivation X =>+ a ... . Both tables are of type SymSetTable.

   The nullable table stores, for each nonterminal, a flag denoting whether
   the nonterminal is nullable (i.e. may be derived to the empty string).

   These tables are constructed by the routines in the YaccClosure unit,
   and are used by the LALR parser construction algorithms in YaccLR0 and
   YaccLookaheads. *)

{=================================================================================}

{$ifdef LargeObjTables}
  {a redimensionable array with elements of type YaccBase.IntSetPtr }
  TDynLimArray_IntSetPtr = class(TDynLimArrayRes)
     protected
       procedure SetElement( Index: longint; value: IntSetPtr );
       function GetElement( Index: longint ): IntSetPtr;
     public
       property Element[ Index: longint ]: IntSetPtr read GetElement write SetElement; default;
       constructor Create;
     end;
{$else}
SymSetTable = array [1..max_nts] of IntSetPtr;
{$endif}

{$ifdef LargeObjTables}
{TDynLimArray_Boolean in YaccTypes}
{$else}
NullableTable = array [1..max_nts] of Boolean;
{$endif}

(* State table:

   Each state stores the first and last index of the kernel items,
   transitions and reductions belonging to it, and a hash key determined
   from the kernel items which is used to speed up searches for existing
   states.

   The items table stores the individual kernel items in the LR(0) set.
   Each entry consists of a rule number together with the item position,
   and a "next" field indicating the associated item in the successor state
   (0 if there is none). The ItemSet type is used to retrieve and manipulate
   individual item sets from the item table.

   The transition table stores the shift and goto transitions in each state
   (each transition is denoted by a (symbol, next_state) pair). Similarly,
   the reductions table stores the reductions in each state, where each
   action is denoted by a (symbolset, ruleno) pair. *)

StateRec = record
             item_lo, item_hi : Integer;
             trans_lo, trans_hi : Integer;
             redns_lo, redns_hi : Integer;
             key : Integer;
           end;

{=================================================================================}

{$ifdef LargeObjTables}
  {a redimensionable array with elements of type YaccTables.StateRec }
  TDynLimArray_StateRec = class(TDynLimArrayRes)
     protected
       procedure SetElement( Index: longint; value: StateRec );
       function GetElement( Index: longint ): StateRec;
     public
       property Element[ Index: longint ]: StateRec read GetElement write SetElement; default;
       constructor Create;
     end;
{$else}
StateTable = array [0..max_states-1] of StateRec;
{$endif}

ItemRec = record
            rule_no, pos_no : Integer;
            next : Integer;
          end;

ItemSet = record
            n_items : Integer;
            item    : array [1..max_set_items] of ItemRec;
          end;

{=================================================================================}

{$ifdef LargeObjTables}
  {a redimensionable array with elements of type YaccTables.ItemRec }
  TDynLimArray_ItemRec = class(TDynLimArrayRes)
     protected
       procedure SetElement( Index: longint; value: ItemRec );
       function GetElement( Index: longint ): ItemRec;
     public
       property Element[ Index: longint ]: ItemRec read GetElement write SetElement; default;
       constructor Create;
     end;
{$else}
ItemTable = array [1..max_items] of ItemRec;
{$endif}

TransRec = record
             sym, next_state : Integer;
           end;

{=================================================================================}

{$ifdef LargeObjTables}
  {a redimensionable array with elements of type YaccTables.TransRec }
  TDynLimArray_TransRec = class(TDynLimArrayRes)
     protected
       procedure SetElement( Index: longint; value: TransRec );
       function GetElement( Index: longint ): TransRec;
     public
       property Element[ Index: longint ]: TransRec read GetElement write SetElement; default;
       constructor Create;
     end;
{$else}
TransTable = array [1..max_trans] of TransRec;
{$endif}

RednRec = record
            symset : IntSetPtr;
            rule_no : Integer;
          end;

{=================================================================================}

{$ifdef LargeObjTables}
  {a redimensionable array with elements of type YaccTables.RednRec }
  TDynLimArray_RednRec = class(TDynLimArrayRes)
     protected
       procedure SetElement( Index: longint; value: RednRec );
       function GetElement( Index: longint ): RednRec;
     public
       property Element[ Index: longint ]: RednRec read GetElement write SetElement; default;
       constructor Create;
     end;
{$else}
RednTable = array [1..max_redns] of RednRec;
{$endif}

(* Lookaheads table: This table stores, for each kernel item, the
   corresponding LALR(1) lookahead symbol sets. *)

{$ifdef LargeObjTables}
{TDynLimArray_IntSetPtr already defined}
{$else}
LookaheadTable = array [1..max_items] of IntSetPtr;
{$endif}

(* The propagation table is used to keep track of how lookaheads are
   propagated from kernel items to other lookahead sets. *)

PropList = ^PropEntry;

PropEntry = record
              symset : IntSetPtr;
              next : PropList;
            end;

{=================================================================================}

{$ifdef LargeObjTables}
  {a redimensionable array with elements of type YaccTables.PropList }
  TDynLimArray_PropList = class(TDynLimArrayRes)
     protected
       procedure SetElement( Index: longint; value: PropList );
       function GetElement( Index: longint ): PropList;
     public
       property Element[ Index: longint ]: PropList read GetElement write SetElement; default;
       constructor Create;
     end;
{$else}
PropTable = array [1..max_items] of PropList;
{$endif}


var
{RZ: verbose and debug switches are defined in yaccbase}
startnt           : Integer;          (* start nonterminal of grammar
                                         (0 if undefined) *)
{$ifdef LargeObjTables}
sym_table         : TDynLimArray_SymRec     ; (* symbol table *)
sym_key           : TDynLimArray_Integer    ; (* symbol keys *)
rule_table        : TDynLimArray_RuleRecPtr ; (* rule table *)
type_table        : TDynLimArray_Integer    ; (* type table *)
sym_type          : TDynLimArray_Integer    ; (* symbol types *)
prec_table        : TDynLimArray_PrecType   ; (* precedence table *)
sym_prec          : TDynLimArray_Integer    ; (* literal symbols precedence *)
rule_prec         : TDynLimArray_Integer    ; (* rules precedence *)
rule_no           : TDynLimArray_Integer    ; (* rule no table *)
rule_offs         : TDynLimArray_RuleOffsRec; (* rule offset table *)
closure_table     : TDynLimArray_IntSetPtr  ; (* closure table *)
first_set_table   : TDynLimArray_IntSetPtr  ; (* first set table *)
nullable          : TDynLimArray_Boolean    ; (* nullable flags table *)
state_table       : TDynLimArray_StateRec   ; (* LR(0) state table *)
item_table        : TDynLimArray_ItemRec    ; (* LR(0) kernel item table *)
trans_table       : TDynLimArray_TransRec   ; (* transition table *)
redn_table        : TDynLimArray_RednRec    ; (* reduction table *)
lookahead_table   : TDynLimArray_IntSetPtr  ; (* LALR lookaheads table *)
prop_table        : TDynLimArray_PropList   ; (* lookahead propagation table *)
{$else}
sym_table         : ^SymTable;      (* symbol table *)
sym_key           : ^SymKeyTable;   (* symbol keys *)
rule_table        : ^RuleTable;     (* rule table *)
type_table        : ^TypeTable;     (* type table *)
sym_type          : ^SymTypeTable;  (* symbol types *)
prec_table        : ^PrecTable;     (* precedence table *)
sym_prec          : ^SymPrecTable;  (* literal symbols precedence *)
rule_prec         : ^RulePrecTable; (* rules precedence *)
rule_no           : ^RuleNoTable;   (* rule no table *)
rule_offs         : ^RuleOffsTable; (* rule offset table *)
closure_table     : ^SymSetTable;   (* closure table *)
first_set_table   : ^SymSetTable;   (* first set table *)
nullable          : ^NullableTable; (* nullable flags table *)
state_table       : ^StateTable;    (* LR(0) state table *)
item_table        : ^ItemTable;     (* LR(0) kernel item table *)
trans_table       : ^TransTable;    (* transition table *)
redn_table        : ^RednTable;     (* reduction table *)
lookahead_table   : ^LookaheadTable;(* LALR lookaheads table *)
prop_table        : ^PropTable;     (* lookahead propagation table *)
{$endif}

(* Operations: *)

(* Symbol table routines: *)

function new_nt : Integer;
  (* returns a new nonterminal number (<-1) *)

function new_lit : Integer;
  (* returns a new literal number above 256 *)

procedure add_lit ( sym : Integer );
  (* this routine allows to add a user-defined literal symbol;
     the current literal symbols count is adjusted accordingly *)

function get_key ( symbol : String ) : Integer;
  (* returns a hash key for symbol *)

procedure def_key ( k : Integer; sym : Integer );
  (* defines k to be a new symbol with internal symbol number sym *)

function is_def_key ( k : Integer; var sym : Integer ) : Boolean;
  (* checks whether symbol denoted by symbol table key k is already
     defined; if so, returns the corresponding symbol number *)

function pname ( sym : Integer ) : String;
  (* returns the print name of an internal symbol (`$end' for
     symbol 0, `$accept' for nonterminal -1, and a single quoted
     character for literals 1..255) *)

(* Rule table routines: *)

function newRuleRec ( r : RuleRec ) : RuleRecPtr;
  (* obtains a dynamic copy of r (only the number of bytes actually
     needed is allocated) *)

procedure add_rule ( r : RuleRecPtr );
  (* add a rule to the rule table *)

procedure sort_rules;
  (* sorts rules w.r.t. left-hand sides into the rule no table *)

procedure rule_offsets;
  (* computes rule offsets after rules have been sorted *)

function n_nt_rules ( sym : Integer ) : Integer;
  (* returns number of rules for nonterminal sym *)

(* Type Table routines: *)

procedure add_type ( k : Integer );
  (* add a type identifier to the table *)

procedure sort_types;
  (* sort the type table alphabetically, eliminate dups *)

function search_type ( symbol : String ) : Boolean;
  (* search the sorted types table for the given type symbol *)

(* Precedence table routines: *)

function new_prec_level ( prec_type : PrecType ) : Integer;
  (* adds a new precedence level of the denoted type; returns: the new
     level *)

(* State table routines: *)

var act_state : Integer; (* state currently considered *)

procedure new_state;
  (* build a new state *)

procedure add_item ( rule_no, pos_no : Integer );
  (* add an item to the new state (initialize its next field to 0) *)

function add_state : Integer;
  (* add the new state to the state table; if an equivalent state is already
     in the table, dispose the new state, and return the existing state
     number, otherwise return the new state number *)

procedure start_trans;
  (* starts building transitions of the active state *)

procedure add_trans ( sym, next_state : Integer );
  (* adds a transition to the active state *)

procedure end_trans;
  (* ends transitions of the active state *)

procedure start_redns;
  (* starts building reduction actions of the active state *)

procedure add_redn ( symset : IntSetPtr; rule_no : Integer );
  (* adds a reduction to the active state *)

procedure end_redns;
  (* ends reduction actions of the active state *)

function n_state_items ( s : Integer ) : Integer;
function n_state_trans ( s : Integer ) : Integer;
function n_state_redns ( s : Integer ) : Integer;
  (* return the number of kernel items, transitions and reductions in state
     s, respectively *)

function find_item( s : Integer; rule_no, pos_no : Integer ) : Integer;
  (* find item (rule_no, pos_no) in state s; returns: the item number *)

(* Item set routines: *)

procedure empty_item_set ( var item_set : ItemSet );
  (* initializes an empty item set *)

procedure include_item_set ( var item_set : ItemSet;
                             rule_no, pos_no : Integer);
  (* add the denoted item to the given item set *)

procedure get_item_set ( s : Integer; var item_set : ItemSet);
  (* obtain the item set of state s from the item table *)

procedure closure ( var item_set : ItemSet );
  (* compute the closure of item_set (using the closure table) *)

procedure sort_item_set ( var item_set : ItemSet );
  (* sorts an item set w.r.t. position and rule numbers (higher positions,
     lower rules first) *)


implementation

uses SysUtils, YaccMsgs, YaccLib,
     {$ifdef VER80} YaccDlg16 {$else} {$ifdef WIN32} YaccDlg32 {$endif} {$endif}
     ;

function n_bytes : LongInt;
  begin
{$ifdef VER80}
    n_bytes := max_bytes-memAvail;
{$else} {$ifdef WIN32}
    n_bytes := AllocMemSize;
{$else}
    n_bytes := 0;
{$endif} {$endif}
  end(*n_bytes*);

(* Symbol table routines: *)

function new_nt : Integer;
  begin
    inc(n_nts);
    if n_nts>max_nts then fatal(nt_table_overflow);
    sym_type{$ifndef LargeObjTables}^{$endif}[-n_nts] := 0;
    new_nt := -n_nts;
  end(*new_nt*);

function new_lit : Integer;
  begin
    inc(n_lits);
    if n_lits>max_lits then fatal(lit_table_overflow);
    sym_type{$ifndef LargeObjTables}^{$endif}[n_lits-1] := 0;
    sym_prec{$ifndef LargeObjTables}^{$endif}[n_lits-1] := 0;
    new_lit := n_lits-1;
  end(*new_lit*);

procedure add_lit ( sym : Integer );
  begin
    if sym>n_lits then n_lits := sym;
    if n_lits>max_lits then fatal(lit_table_overflow);
    sym_type{$ifndef LargeObjTables}^{$endif}[sym] := 0;
    sym_prec{$ifndef LargeObjTables}^{$endif}[sym] := 0;
  end(*add_lit*);

{$F+}
function lookup(k : Integer) : String;
{$F-}
  (* print name of symbol no. k *)
  begin
    with sym_table{$ifndef LargeObjTables}^{$endif}[k] do
      if pname=nil then
        lookup := ''
      else
        lookup := pname^
  end(*lookup*);

{$F+}
procedure entry(k : Integer; symbol : String);
{$F-}
{$ifdef LargeObjTables}
  var tmp: SymRec;
{$endif}
  (* enter symbol into table *)
  begin
{$ifdef LargeObjTables}
    tmp := sym_table[k]; tmp.pname := newStr(symbol); sym_table[k] := tmp;
{$else}
    sym_table^[k].pname := newStr(symbol);
{$endif}
  end(*entry*);

function get_key ( symbol : String ) : Integer;
  begin
    get_key := key(symbol, max_keys, lookup, entry);
  end(*get_key*);

procedure def_key ( k : Integer; sym : Integer );
{$ifdef LargeObjTables}
  var tmp: SymRec;
{$endif}
  begin
    sym_key{$ifndef LargeObjTables}^{$endif}[sym] := k;
{$ifdef LargeObjTables}
    tmp := sym_table[k]; tmp.deff := true;
    tmp.sym  := sym; sym_table[k] := tmp;
{$else}
    sym_table^[k].deff := true;
    sym_table^[k].sym  := sym;
{$endif}
  end(*def_key*);

function is_def_key ( k : Integer; var sym : Integer ) : Boolean;
  begin
    if sym_table{$ifndef LargeObjTables}^{$endif}[k].deff then
      begin
        sym := sym_table{$ifndef LargeObjTables}^{$endif}[k].sym;
        is_def_key := true;
      end
    else
      is_def_key := false
  end(*is_def_key*);

function pname ( sym : Integer ) : String;
  begin
    case sym of
      1..255 : pname := singleQuoteStr(chr(sym));
      0      : pname := '$end';
      -1     : pname := '$accept';
      else if sym_table{$ifndef LargeObjTables}^{$endif}[
                sym_key{$ifndef LargeObjTables}^{$endif}[sym] ].pname^[1]='''' then
        pname := singleQuoteStr(
                   copy( sym_table{$ifndef LargeObjTables}^{$endif}[
                           sym_key{$ifndef LargeObjTables}^{$endif}[sym] ].pname^,
                         2,
                         length(sym_table{$ifndef LargeObjTables}^{$endif}[
                                  sym_key{$ifndef LargeObjTables}^{$endif}[sym] ].pname^)-2)
                 )
      else
        pname := sym_table{$ifndef LargeObjTables}^{$endif}[
                   sym_key{$ifndef LargeObjTables}^{$endif}[sym] ].pname^
    end;
  end(*pname*);

(* Rule table: *)

function newRuleRec ( r : RuleRec ) : RuleRecPtr;
  var rp : RuleRecPtr;
  begin
    getmem(rp, 2*sizeOf(Integer)+r.rhs_len*sizeOf(Integer));
    move(r, rp^, 2*sizeOf(Integer)+r.rhs_len*sizeOf(Integer));
    newRuleRec := rp;
  end(*newRuleRec*);

procedure add_rule ( r : RuleRecPtr );
  begin
    inc(n_rules);
    if n_rules>max_rules then fatal(rule_table_overflow);
    rule_table{$ifndef LargeObjTables}^{$endif}[n_rules] := r;
  end(*add_rule*);

{$F+}
function rule_less ( i, j : Integer ) : Boolean;
{$F-}
  begin
    if rule_table{$ifndef LargeObjTables}^{$endif}[
         rule_no{$ifndef LargeObjTables}^{$endif}[i] ]^.lhs_sym =
       rule_table{$ifndef LargeObjTables}^{$endif}[
         rule_no{$ifndef LargeObjTables}^{$endif}[j] ]^.lhs_sym then
      rule_less := rule_no{$ifndef LargeObjTables}^{$endif}[i] <
                     rule_no{$ifndef LargeObjTables}^{$endif}[j]
    else
      rule_less := rule_table{$ifndef LargeObjTables}^{$endif}[
                     rule_no{$ifndef LargeObjTables}^{$endif}[i] ]^.lhs_sym >
                   rule_table{$ifndef LargeObjTables}^{$endif}[
                     rule_no{$ifndef LargeObjTables}^{$endif}[j] ]^.lhs_sym
  end(*rule_less*);

{$F+}
procedure rule_swap ( i, j : Integer );
{$F-}
  var x : Integer;
  begin
    x := rule_no{$ifndef LargeObjTables}^{$endif}[i];
    rule_no{$ifndef LargeObjTables}^{$endif}[i] := rule_no{$ifndef LargeObjTables}^{$endif}[j];
    rule_no{$ifndef LargeObjTables}^{$endif}[j] := x;
  end(*rule_swap*);

procedure sort_rules;
  var i : Integer;
  begin
    for i := 1 to n_rules do rule_no{$ifndef LargeObjTables}^{$endif}[i] := i;
    quicksort ( 1, n_rules, rule_less, rule_swap );
  end(*sort_rules*);

procedure rule_offsets;
  var i, sym : Integer;
{$ifdef LargeObjTables}
      tmp : RuleOffsRec;
{$endif}
  begin
    YaccMsgs.beep;
    for sym := 1 to n_nts do begin
      if (sym mod 30 = 0 ) then YaccMsgs.beep;
{$ifdef LargeObjTables}
      tmp := rule_offs[sym];
      with tmp do
        begin
          rule_lo := 1; rule_hi := 0;
        end;
      rule_offs[sym] := tmp;
{$else}
      with rule_offs^[sym] do
        begin
          rule_lo := 1; rule_hi := 0;
        end;
{$endif}
    end;
    i := 1;
    while (i<=n_rules) do
      begin
        sym := rule_table{$ifndef LargeObjTables}^{$endif}[
                 rule_no{$ifndef LargeObjTables}^{$endif}[i] ]^.lhs_sym;
{$ifdef LargeObjTables}
        tmp := rule_offs[-sym]; tmp.rule_lo := i; rule_offs[-sym] := tmp;
{$else}
        rule_offs^[-sym].rule_lo := i;
{$endif}
        if (i mod 20 = 0 ) then YaccMsgs.beep;
        inc(i);  {$B-}
        while (i<=n_rules) and
              (rule_table{$ifndef LargeObjTables}^{$endif}[
                 rule_no{$ifndef LargeObjTables}^{$endif}[i] ]^.lhs_sym=sym) do
          inc(i);
{$ifdef LargeObjTables}
        tmp := rule_offs[-sym]; tmp.rule_hi := i-1; rule_offs[-sym] := tmp;
{$else}
        rule_offs^[-sym].rule_hi := i-1;
{$endif}
      end;
  end(*rule_offsets*);

function n_nt_rules ( sym : Integer ) : Integer;
  begin
    with rule_offs{$ifndef LargeObjTables}^{$endif}[-sym] do
      n_nt_rules := rule_hi-rule_lo+1;
        {RZ: Bug fixed; thank goodness, this routine was not used at all}
  end(*n_nt_rules*);

(* Type Table routines: *)

procedure add_type ( k : Integer );
  begin
    inc(n_types);
    if n_types>max_types then fatal(type_table_overflow);
    type_table{$ifndef LargeObjTables}^{$endif}[n_types] := k;
  end(*add_type*);

(* Routines to sort type identifiers alphabetically: *)

{$F+}
function type_less ( i, j : Integer ) : Boolean;
{$F-}
  begin
    type_less := sym_table{$ifndef LargeObjTables}^{$endif}[
                   type_table{$ifndef LargeObjTables}^{$endif}[i] ].pname^<
                 sym_table{$ifndef LargeObjTables}^{$endif}[
                   type_table{$ifndef LargeObjTables}^{$endif}[j] ].pname^
  end(*type_less*);

{$F+}
procedure type_swap ( i, j : Integer );
{$F-}
  var x : Integer;
  begin
    x := type_table{$ifndef LargeObjTables}^{$endif}[i];
    type_table{$ifndef LargeObjTables}^{$endif}[i] := type_table{$ifndef LargeObjTables}^{$endif}[j];
    type_table{$ifndef LargeObjTables}^{$endif}[j] := x;
  end(*type_swap*);

procedure sort_types;
  var i, j, count : Integer; {RZ: sym was not used here, removed it}
  begin
    (* sort: *)
    quicksort(1, n_types, type_less, type_swap);
    (* eliminate dups: *)
    i := 1; j := 1; count := 0;
    while i<=n_types do
      begin
        if i<>j then type_table{$ifndef LargeObjTables}^{$endif}[j] :=
                       type_table{$ifndef LargeObjTables}^{$endif}[i];
        while (i<n_types) and
              ( type_table{$ifndef LargeObjTables}^{$endif}[i+1] =
                  type_table{$ifndef LargeObjTables}^{$endif}[i] ) do
          begin
            inc(i); inc(count);
          end;
        inc(i); inc(j);
      end;
    dec(n_types, count);
  end(*sort_types*);

function search_type ( symbol : String ) : Boolean;
  var l, r, k : Integer;
  begin
    (* binary search: *)
    l := 1; r := n_types;
    k := l + (r-l) div 2;
    while (l<r) and (sym_table{$ifndef LargeObjTables}^{$endif}[
                       type_table{$ifndef LargeObjTables}^{$endif}[k] ].pname^<>symbol) do
      begin
        if sym_table{$ifndef LargeObjTables}^{$endif}[
             type_table{$ifndef LargeObjTables}^{$endif}[k] ].pname^<symbol then
          l := succ(k)
        else
          r := pred(k);
        k := l + (r-l) div 2;
      end;
    search_type := (k<=n_types) and (sym_table{$ifndef LargeObjTables}^{$endif}[
                             type_table{$ifndef LargeObjTables}^{$endif}[k] ].pname^=symbol);
  end(*search_type*);

(* Precedence table routines: *)

function new_prec_level ( prec_type : PrecType ) : Integer;
  begin
    inc(n_prec);
    if n_prec>max_prec then fatal(prec_table_overflow);
    prec_table{$ifndef LargeObjTables}^{$endif}[n_prec] := prec_type;
    new_prec_level := n_prec;
  end(*new_prec_level*);

(* State table: *)

procedure new_state;
{$ifdef LargeObjTables}
  var tmp : StateRec;
{$endif}
  begin
    inc(n_states);
    if n_states>max_states then fatal(state_table_overflow);
{$ifdef LargeObjTables}
    tmp := state_table[n_states-1]; tmp.item_lo := n_items+1; state_table[n_states-1] := tmp;
{$else}
    state_table^[n_states-1].item_lo := n_items+1;
{$endif}
  end(*new_state*);

procedure add_item ( rule_no, pos_no : Integer );
{$ifdef LargeObjTables}
  var tmp : ItemRec;
{$endif}
  begin
    inc(n_items);
    if n_items>max_items then fatal(item_table_overflow);
{$ifdef LargeObjTables}
    tmp := item_table[n_items];
    tmp.rule_no := rule_no;
    tmp.pos_no  := pos_no;
    tmp.next    := 0;
    item_table[n_items] := tmp;
{$else}
    item_table^[n_items].rule_no := rule_no;
    item_table^[n_items].pos_no  := pos_no;
    item_table^[n_items].next    := 0;
{$endif}
  end(*add_item*);

function add_state : Integer;
  function state_key ( s : Integer ) : Integer;
    (* determines a hash key for state s *)
    const max_key = 4001;
      (* should be prime number s.t. hash keys are distributed
         evenly *)
      {RZ: different from max_keys; for a prime see Prime-table above}
    var i : Integer;
        k : longint;
    begin
      with state_table{$ifndef LargeObjTables}^{$endif}[s] do
        begin
          k := 0;
          for i := item_lo to item_hi do
            with item_table{$ifndef LargeObjTables}^{$endif}[i] do
              inc(k, rule_no+pos_no);
          state_key := k mod max_key;
        end;
    end(*state_key*);
  function search_state ( s, lo, hi : Integer; var t : Integer ) : Boolean;
    (* searches the range lo..hi in the state table for a state with the
       same kernel items as s; returns true if found, and then the
       corresponding state number in t *)
    function eq_items(s, t : Integer) : Boolean;
      (* compares kernel item sets of states s and t *)
      var i, i_s, i_t : Integer;
      begin
        if n_state_items(s)<>n_state_items(t) then
          eq_items := false
        else
          begin
            i_s := state_table{$ifndef LargeObjTables}^{$endif}[s].item_lo;
            i_t := state_table{$ifndef LargeObjTables}^{$endif}[t].item_lo;
            for i := 0 to n_state_items(s)-1 do
              if ( item_table{$ifndef LargeObjTables}^{$endif}[i_s+i].rule_no <>
                     item_table{$ifndef LargeObjTables}^{$endif}[i_t+i].rule_no  ) or
                 ( item_table{$ifndef LargeObjTables}^{$endif}[i_s+i].pos_no <>
                     item_table{$ifndef LargeObjTables}^{$endif}[i_t+i].pos_no   ) then
                begin
                  eq_items := false;
                  exit;
                end;
            eq_items := true;
          end
      end(*eq_items*);
    var t1 : Integer;
    begin
      with state_table{$ifndef LargeObjTables}^{$endif}[s] do
        for t1 := lo to hi do
          if (key=state_table{$ifndef LargeObjTables}^{$endif}[t1].key) and
             eq_items(s, t1) then
            begin
              search_state := true;
              t := t1;
              exit;
            end;
      search_state := false;
    end(*search_state*);
  var s : Integer;
{$ifdef LargeObjTables}
      tmp : StateRec; tmp_i : longint;
{$endif}
  begin
{$ifdef LargeObjTables}
    tmp_i := n_states-1;
    tmp := state_table[tmp_i];
    tmp.item_hi := n_items;
      {RZ: necessary, as func state_key uses all elems of state_table including this one!}
    state_table[tmp_i] := tmp;
    {RZ: the following function uses contents of tmp resp. state_table[n_states-1]}
    tmp.key := state_key(n_states-1);  state_table[tmp_i] := tmp;
    if search_state(n_states-1, 0, n_states-2, s) then
      begin
        n_items := state_table[n_states-1].item_lo;
        dec(n_states);
        add_state := s;
      end
    else
      add_state := n_states-1;
{$else}
    with state_table^[n_states-1] do
      begin
        item_hi := n_items;
        key := state_key(n_states-1);
        if search_state(n_states-1, 0, n_states-2, s) then
          begin
            n_items := item_lo;
            dec(n_states);
            add_state := s;
          end
        else
          add_state := n_states-1;
      end;
{$endif}
  end(*add_state*);

procedure start_trans;
{$ifdef LargeObjTables}
  var tmp : StateRec;
{$endif}
  begin
{$ifdef LargeObjTables}
    tmp := state_table[act_state]; tmp.trans_lo := n_trans+1; state_table[act_state] := tmp;
{$else}
    state_table^[act_state].trans_lo := n_trans+1;
{$endif}
  end(*start_trans*);

procedure add_trans ( sym, next_state : Integer );
{$ifdef LargeObjTables}
  var tmp : TransRec;
{$endif}
  begin
    inc(n_trans);
    if n_trans>max_trans then fatal(trans_table_overflow);
{$ifdef LargeObjTables}
    tmp := trans_table[n_trans]; tmp.sym := sym;
    tmp.next_state := next_state; trans_table[n_trans] := tmp;
{$else}
    trans_table^[n_trans].sym        := sym;
    trans_table^[n_trans].next_state := next_state;
{$endif}
  end(*add_trans*);

procedure end_trans;
{$ifdef LargeObjTables}
  var tmp : StateRec;
{$endif}
  begin
{$ifdef LargeObjTables}
    tmp := state_table[act_state]; tmp.trans_hi := n_trans; state_table[act_state] := tmp;
{$else}
    state_table^[act_state].trans_hi := n_trans;
{$endif}
  end(*end_trans*);

procedure start_redns;
{$ifdef LargeObjTables}
  var tmp : StateRec;
{$endif}
  begin
{$ifdef LargeObjTables}
    tmp := state_table[act_state]; tmp.redns_lo := n_redns+1; state_table[act_state] := tmp;
{$else}
    state_table^[act_state].redns_lo := n_redns+1;
{$endif}
  end(*start_redns*);

procedure add_redn ( symset : IntSetPtr; rule_no : Integer );
{$ifdef LargeObjTables}
  var tmp : RednRec;
{$endif}
  begin
    inc(n_redns);
    if n_redns>max_redns then fatal(redn_table_overflow);
{$ifdef LargeObjTables}
    tmp := redn_table[n_redns];
    tmp.symset  := symset;
    tmp.rule_no := rule_no;
    redn_table[n_redns] := tmp;
{$else}
    redn_table^[n_redns].symset  := symset;
    redn_table^[n_redns].rule_no := rule_no;
{$endif}
  end(*add_redn*);

procedure end_redns;
{$ifdef LargeObjTables}
  var tmp : StateRec;
{$endif}
  begin
{$ifdef LargeObjTables}
    tmp := state_table[act_state]; tmp.redns_hi := n_redns; state_table[act_state] := tmp;
{$else}
    state_table^[act_state].redns_hi := n_redns;
{$endif}
  end(*end_redns*);

function n_state_items ( s : Integer ) : Integer;
  begin
    with state_table{$ifndef LargeObjTables}^{$endif}[s] do
      n_state_items := item_hi-item_lo+1
  end(*n_state_items*);

function n_state_trans ( s : Integer ) : Integer;
  begin
    with state_table{$ifndef LargeObjTables}^{$endif}[s] do
      n_state_trans := trans_hi-trans_lo+1
  end(*n_state_trans*);

function n_state_redns ( s : Integer ) : Integer;
  begin
    with state_table{$ifndef LargeObjTables}^{$endif}[s] do
      n_state_redns := redns_hi-redns_lo+1
  end(*n_state_redns*);

function find_item( s : Integer; rule_no, pos_no : Integer ) : Integer;
  var i : Integer;
  begin
    with state_table{$ifndef LargeObjTables}^{$endif}[s] do
      for i := item_lo to item_hi do
        if (item_table{$ifndef LargeObjTables}^{$endif}[i].rule_no=rule_no) and
           (item_table{$ifndef LargeObjTables}^{$endif}[i].pos_no=pos_no) then
          begin
            find_item := i;
            exit;
          end;
    find_item := 0;
  end(*find_item*);

(* Item set routines: *)

procedure empty_item_set ( var item_set : ItemSet );
  begin
    item_set.n_items := 0;
  end(*empty_item_set*);

procedure include_item_set ( var item_set : ItemSet;
                             rule_no, pos_no : Integer);
  begin
    with item_set do
      begin
        inc(n_items);
        if n_items> max_set_items then fatal(item_table_overflow+' (set)');
        item[n_items].rule_no := rule_no;
        item[n_items].pos_no  := pos_no;
      end;
  end(*include_item_set*);

procedure get_item_set ( s : Integer; var item_set : ItemSet);
  begin
    with state_table{$ifndef LargeObjTables}^{$endif}[s], item_set do
      begin
        n_items := n_state_items(s);
{$ifdef LargeObjTables}
        move(item_table.GetElementPtr(item_lo)^, item, n_items*sizeOf(ItemRec));
{$else}
        move(item_table^[item_lo], item, n_items*sizeOf(ItemRec));
{$endif}
      end
  end(*get_item_set*);

procedure closure ( var item_set : ItemSet );
  var i, j : Integer;
      nt_syms0, nt_syms : IntSet;
  begin
    with item_set do
      begin
        (* get the nonterminals at current positions in items: *)
        empty(nt_syms0);
        for i := 1 to n_items do
          with item[i], rule_table{$ifndef LargeObjTables}^{$endif}[rule_no]^ do
            if (pos_no<=rhs_len) and (rhs_sym[pos_no]<0) then
              include(nt_syms0, rhs_sym[pos_no]);
        nt_syms := nt_syms0;
        (* add closure symbols: *)
        for i := 1 to size(nt_syms0) do
          setunion(nt_syms, closure_table{$ifndef LargeObjTables}^{$endif}[-nt_syms0[i]]^);
        (* add the nonkernel items for the nonterminal symbols: *)
        for i := 1 to size(nt_syms) do
          with rule_offs{$ifndef LargeObjTables}^{$endif}[-nt_syms[i]] do
            for j := rule_lo to rule_hi do
              include_item_set(item_set, rule_no{$ifndef LargeObjTables}^{$endif}[j], 1);
      end;
  end(*closure*);

var sort_items : ItemSet;

(* comparison and swap routines for sort_item_set: *)

{$F+}
function items_less ( i, j : Integer ) : Boolean;
{$F-}
  begin
    with sort_items do
      if item[i].pos_no=item[j].pos_no then
        items_less := item[i].rule_no<item[j].rule_no
      else
        items_less := item[i].pos_no>item[j].pos_no
  end(*items_less*);

{$F+}
procedure items_swap ( i, j : Integer );
{$F-}
  var x : ItemRec;
  begin
    with sort_items do
      begin
        x := item[i]; item[i] := item[j]; item[j] := x;
      end
  end(*items_swap*);

procedure sort_item_set ( var item_set : ItemSet );
  begin
    sort_items := item_set;
    quicksort(1, sort_items.n_items, items_less, items_swap);
    item_set := sort_items;
  end(*sort_item_set*);


{$ifdef LargeObjTables}
{=================================================================================}

constructor TDynLimArray_SymRec.Create;
begin
  inherited Create;
  ElementSize := System.Sizeof(SymRec);
end;

procedure TDynLimArray_SymRec.SetElement( Index: longint; value: SymRec );
var temp : ^SymRec;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_Write);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    temp^ := value;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

function TDynLimArray_SymRec.GetElement( Index: longint ): SymRec;
var temp : ^SymRec;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_read);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    result := temp^;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

{=================================================================================}

constructor TDynLimArray_RuleRecPtr.Create;
begin
  inherited Create;
  ElementSize := System.Sizeof(RuleRecPtr);
end;

procedure TDynLimArray_RuleRecPtr.SetElement( Index: longint; value: RuleRecPtr );
var temp : ^RuleRecPtr;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_Write);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    temp^ := value;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

function TDynLimArray_RuleRecPtr.GetElement( Index: longint ): RuleRecPtr;
var temp : ^RuleRecPtr;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_read);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    result := temp^;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

{=================================================================================}

constructor TDynLimArray_RuleOffsRec.Create;
begin
  inherited Create;
  ElementSize := System.Sizeof(RuleOffsRec);
end;

procedure TDynLimArray_RuleOffsRec.SetElement( Index: longint; value: RuleOffsRec );
var temp : ^RuleOffsRec;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_Write);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    temp^ := value;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

function TDynLimArray_RuleOffsRec.GetElement( Index: longint ): RuleOffsRec;
var temp : ^RuleOffsRec;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_read);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    result := temp^;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

{=================================================================================}

constructor TDynLimArray_PrecType.Create;
begin
  inherited Create;
  ElementSize := System.Sizeof(PrecType);
end;

procedure TDynLimArray_PrecType.SetElement( Index: longint; value: PrecType );
var temp : ^PrecType;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_Write);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    temp^ := value;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

function TDynLimArray_PrecType.GetElement( Index: longint ): PrecType;
var temp : ^PrecType;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_read);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    result := temp^;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

{=================================================================================}

constructor TDynLimArray_IntSetPtr.Create;
begin
  inherited Create;
  ElementSize := System.Sizeof(IntSetPtr);
end;

procedure TDynLimArray_IntSetPtr.SetElement( Index: longint; value: IntSetPtr );
var temp : ^IntSetPtr;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_Write);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    temp^ := value;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

function TDynLimArray_IntSetPtr.GetElement( Index: longint ): IntSetPtr;
var temp : ^IntSetPtr;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_read);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    result := temp^;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

{=================================================================================}

constructor TDynLimArray_StateRec.Create;
begin
  inherited Create;
  ElementSize := System.Sizeof(StateRec);
end;

procedure TDynLimArray_StateRec.SetElement( Index: longint; value: StateRec );
var temp : ^StateRec;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_Write);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    temp^ := value;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

function TDynLimArray_StateRec.GetElement( Index: longint ): StateRec;
var temp : ^StateRec;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_read);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    result := temp^;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

{=================================================================================}

constructor TDynLimArray_ItemRec.Create;
begin
  inherited Create;
  ElementSize := System.Sizeof(ItemRec);
end;

procedure TDynLimArray_ItemRec.SetElement( Index: longint; value: ItemRec );
var temp : ^ItemRec;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_Write);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    temp^ := value;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

function TDynLimArray_ItemRec.GetElement( Index: longint ): ItemRec;
var temp : ^ItemRec;
begin
  if (Index > HighIndex) or (Index < LowIndex) then begin
    raise EDynLimArrayConcrete.Create(Index_Error_read);
  end;
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    result := temp^;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

{=================================================================================}

constructor TDynLimArray_TransRec.Create;
begin
  inherited Create;
  ElementSize := System.Sizeof(TransRec);
end;

procedure TDynLimArray_TransRec.SetElement( Index: longint; value: TransRec );
var temp : ^TransRec;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_Write);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    temp^ := value;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

function TDynLimArray_TransRec.GetElement( Index: longint ): TransRec;
var temp : ^TransRec;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_read);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    result := temp^;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

{=================================================================================}

constructor TDynLimArray_RednRec.Create;
begin
  inherited Create;
  ElementSize := System.Sizeof(RednRec);
end;

procedure TDynLimArray_RednRec.SetElement( Index: longint; value: RednRec );
var temp : ^RednRec;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_Write);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    temp^ := value;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

function TDynLimArray_RednRec.GetElement( Index: longint ): RednRec;
var temp : ^RednRec;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_read);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    result := temp^;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

{=================================================================================}

constructor TDynLimArray_PropList.Create;
begin
  inherited Create;
  ElementSize := System.Sizeof(PropList);
end;

procedure TDynLimArray_PropList.SetElement( Index: longint; value: PropList );
var temp : ^PropList;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_Write);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    temp^ := value;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

function TDynLimArray_PropList.GetElement( Index: longint ): PropList;
var temp : ^PropList;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_read);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    result := temp^;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

{$endif}
{=================================================================================}

var i : Integer;
{$ifdef LargeObjTables}
    tmp : SymRec;
{$endif}
initialization

  verbose          := false;
  debug            := false;
  startnt          := 0;

{$ifdef VER80}
  max_bytes := memAvail;
{$else} {$ifdef WIN32}
  max_bytes := AllocMemSize;
{$else}
  max_bytes := 0;
{$endif} {$endif}

  n_nts            := 1;
  n_lits           := 257;
  n_rules          := 0;
  n_types          := 0;
  n_prec           := 0;
  n_states         := 0;
  n_items          := 0;
  n_trans          := 0;
  n_redns          := 0;

  (* allocate tables: *)

{$ifdef LargeObjTables}
    {SymTable = array [1..max_keys] of SymRec;}
  sym_table := TDynLimArray_SymRec.Create;
  sym_table.Dim( 1, max_keys );
    {SymKeyTable = array [-max_nts..max_lits-1] of Integer;}
  sym_key := TDynLimArray_Integer.Create;
  sym_key.Dim( -max_nts, max_lits-1 );
    {RuleTable = array [1..max_rules] of RuleRecPtr;}
  rule_table := TDynLimArray_RuleRecPtr.Create;
  rule_table.Dim( 1, max_rules );
    {RuleNoTable = array [1..max_rules] of Integer;}
  rule_no := TDynLimArray_Integer.Create;
  rule_no.Dim( 1, max_rules );
    {RuleOffsTable = array [1..max_nts] of RuleOffsRec;}
  rule_offs := TDynLimArray_RuleOffsRec.Create;
  rule_offs.Dim( 1, max_nts );
    {TypeTable    = array [1..max_types] of Integer;}
  type_table := TDynLimArray_Integer.Create;
  type_table.Dim( 1, max_types );
    {SymTypeTable = array [-max_nts..max_lits-1] of Integer;}
  sym_type := TDynLimArray_Integer.Create;
  sym_type.Dim( -max_nts, max_lits-1 );
    {PrecTable = array [1..max_prec] of PrecType;}
  prec_table := TDynLimArray_PrecType.Create;
  prec_table.Dim( 1, max_prec );
    {SymPrecTable = array [0..max_lits-1] of Integer;}
  sym_prec := TDynLimArray_Integer.Create;
  sym_prec.Dim( 0, max_lits-1 );
    {RulePrecTable = array [1..max_rules] of Integer;}
  rule_prec := TDynLimArray_Integer.Create;
  rule_prec.Dim( 1, max_rules );
    {SymSetTable = array [1..max_nts] of IntSetPtr;}
  closure_table := TDynLimArray_IntSetPtr.Create;
  closure_table.Dim( 1, max_nts );
    {SymSetTable = array [1..max_nts] of IntSetPtr;}
  first_set_table := TDynLimArray_IntSetPtr.Create;
  first_set_table.Dim( 1, max_nts );
    {NullableTable = array [1..max_nts] of Boolean;}
  nullable := TDynLimArray_Boolean.Create;
  nullable.Dim( 1, max_nts );
    {StateTable = array [0..max_states-1] of StateRec;}
  state_table := TDynLimArray_StateRec.Create;
  state_table.Dim( 0, max_states-1 );
    {ItemTable = array [1..max_items] of ItemRec;}
  item_table := TDynLimArray_ItemRec.Create;
  item_table.Dim( 1, max_items );
    {TransTable = array [1..max_trans] of TransRec;}
  trans_table := TDynLimArray_TransRec.Create;
  trans_table.Dim( 1, max_trans );
    {RednTable = array [1..max_redns] of RednRec;}
  redn_table := TDynLimArray_RednRec.Create;
  redn_table.Dim( 1, max_redns );
    {LookaheadTable = array [1..max_items] of IntSetPtr;}
  lookahead_table := TDynLimArray_IntSetPtr.Create;
  lookahead_table.Dim( 1, max_items );
    {PropTable = array [1..max_items] of PropList;}
  prop_table := TDynLimArray_PropList.Create;
  prop_table.Dim( 1, max_items );
{$else}
  new(sym_table);
  new(sym_key);
  new(rule_table);
  new(rule_no);
  new(rule_offs);
  new(type_table);
  new(sym_type);
  new(prec_table);
  new(sym_prec);
  new(rule_prec);
  new(closure_table);
  new(first_set_table);
  new(nullable);
  new(state_table);
  new(item_table);
  new(trans_table);
  new(redn_table);
  new(lookahead_table);
  new(prop_table);
{$endif}

  (* initialize symbol table: *)

{$ifdef LargeObjTables}
  with tmp do
    begin
    pname := nil;
    deff  := false;
    sym := 0;
    end;
  for i := 1 to max_keys do begin
    sym_table[i] := tmp;
  end;
{$else}
  for i := 1 to max_keys do
    with sym_table^[i] do
    begin
    pname := nil;
    deff  := false;
  end;
{$endif}

  (* enter predefined error symbol into symbol table: *)

  def_key(get_key('error'), 256);

  (* initialize type and precedence tables: *)

  for i := -max_nts to max_lits-1 do sym_type{$ifndef LargeObjTables}^{$endif}[i] := 0;
  for i := 0 to max_lits-1 do sym_prec{$ifndef LargeObjTables}^{$endif}[i] := 0;
  for i := 1 to max_rules do rule_prec{$ifndef LargeObjTables}^{$endif}[i] := 0;

end(*YaccTables*).     
